"""
provides utilities to work with the apis relvant for vehicledata
"""
import pandas as pd
import requests as rq
from .vehicleDataUtils.read import vehicledata_from_dict

vehiclesApi_url = f"https://service.ivanto.de/srs/api/v1/vehiclelivedata?tenant=heag"
stationsApi_url = f"https://service.ivanto.de/srs/api/v1/stations?bundleIdentifier=de.ivanto.heagmobilo&tenant=heag&onlyStations=true&latitude=49.872021&longitude=8.658934&distance=50000"

def get_current_vehicle_data():
    """ request vehicleLiveData form api
    """
    response = rq.get(vehiclesApi_url)

    if response.status_code != rq.codes.ok:
        RuntimeError('Api response status code not ok!')
    
    return response.json()

def get_station_data():
    """ request stationData form api
    """
    response = rq.get(stationsApi_url)

    if response.status_code != rq.codes.ok:
        RuntimeError('Api response status code not ok!')
    
    return response.json()


def get_current_vehicle_data_as_dataframe() -> pd.DataFrame:
    """
    gets current vehiclelivedata from api and  returns it as a dataframe

    Returns: vehicleLiveData form api as pandas dataframe

    """
    return vehicledata_from_dict(get_current_vehicle_data())


import unittest

import pandas as pd
from heagVehicleLivedataUtils.analyze.plot import VehicleDataPlotter
from heagVehicleLivedataUtils.analyze.data import DataAnalysis
from heagVehicleLivedataUtils.vehicleInformation import encode_line_name
from heagVehicleLivedataUtils.vehicleDataUtils.read import verify_vehicledata_format

class TestRead(unittest.TestCase):
    def test_regular_read_and_plot(self):
        da = VehicleDataPlotter(vehicledata_path="../test/vehicleInfo_test/")

        # test if working with data does not throw errors
        da.plot_number_of_trams_in_service(sample_time="15Min",show_plot=False)
        da.plot_all_trams_in_service(sample_time="15Min",show_plot=False)
        da.plot_electric_buses_in_service(sample_time="15Min",show_plot=False)

        da.get_tram_journeys()
        da.get_tram_journeys()

        # check if dataframe is formated to expected spec
        self.assertTrue(verify_vehicledata_format(da.get_vehicledata()))


    # TODO: what about ill formed data?/ -> error handling
    def test_special_read(self):
        da = VehicleDataPlotter(vehicledata_path="../test/vehicleInfo_test_special_cases/") # TODO seems to have problems with "6E" and such lines
        vehicle_data = da.get_vehicledata()

        timestamp = pd.Timestamp("2024-11-09T09:29:49+0100")

        # bus line as category 1(tram)
        self.assertEqual(vehicle_data.loc[(timestamp, 444),'category'],1)

        # added offset to vehicleid
        self.assertEqual(vehicle_data.loc[(timestamp, 4284), 'lineid'], encode_line_name("L"))

        # line name that can be read as float
        self.assertEqual(vehicle_data.loc[(timestamp, 430), 'lineid'], encode_line_name("4E"))

        # tram vehicleid on bus line
        self.assertEqual(vehicle_data.loc[(timestamp, 112), 'lineid'], encode_line_name("WE2"))


        timestamp = pd.Timestamp("2024-11-09T23:04:49+0100")

        # line name that can be read as float
        self.assertEqual(vehicle_data.loc[(timestamp, 69), 'lineid'], encode_line_name("8E"))

        # switched category
        self.assertEqual(vehicle_data.loc[(timestamp, 114), 'category'], 5)


        timestamp = pd.Timestamp("2024-11-09T23:39:49+0100")

        self.assertEqual(vehicle_data.loc[(timestamp, 405), 'lineid'], encode_line_name("8"))

        self.assertEqual(vehicle_data.loc[(timestamp, 68), 'lineid'], encode_line_name("6E"))



if __name__ == '__main__':
    unittest.main()



"""
This module contains constants regarding operations 

TODO: das ordentlich beschreiben, ie. hier sind vorallem betreibsnummber der Fahrzeuge und ne colormap für die liniennummern ihre farbe zuordent
"""

from matplotlib.colors import ListedColormap


def tram_class_prefix(vehicle_id: int) -> str:
    """ gives the class prefix for the vehicle number of the given tram

    Args:
        vehicle_id: the (TODO: vehicleID) of a tram without the class prefix

    Returns:
        string: the class prefix of that tram

    """

    # sort of divide & conquer to avoid elif chains
    if vehicle_id >= 75:
        if vehicle_id <= 92:
            return '07'
        elif vehicle_id <= 125:
            return '22'
        else:
            return ''  # not in range of current numbers
    else:
        if vehicle_id >= 55:
            return '98'
        elif vehicle_id <= 25:
            # modern hight floor tram
            if vehicle_id >= 15:
                return '91'
            elif vehicle_id >= 9:
                return '82'
            else:
                return '76'
        else:
            return ''  # trailers are not shown as vehicles, others have no prefix


def get_tram_number(vehicle_id: int) -> str:
    """ gives the full id number of the tram vehicle

    Args:
        vehicle_id: the vehicle Id number of the tram whose full id number is to be returned.

    Returns:
        str: the full id number of the tram vehicle
    """
    return tram_class_prefix(vehicle_id) + str(vehicle_id)


# the vehicle numbers of the differnet classes of trams
st12Numbers: list[str] = [f'91{i:0>2}' for i in range(15, 25)]
st13Numbers: list[str] = [f'98{i:0>2}' for i in range(55, 75)]
st14Numbers: list[str] = [f'07{i:0>2}' for i in range(75, 93)]
st15Numbers: list[str] = [f'221{i:0>2}' for i in range(1, 26)]
tinaNumbers: list[str] = st15Numbers
st7Numbers: list[str] = ['31']

modern_tram_numbers: list[str] = st12Numbers + st13Numbers + st14Numbers + tinaNumbers
usual_tram_numbers: list[str] = st7Numbers + modern_tram_numbers

# Betriebsnummern of the busses in operation
setra_s531_dt_numbers = ['291','292','298','299'] # double decker (2019-2021)

citaro_c2_o530_le_numbers = [f'31{i:0>1}' for i in range(7,10)] # standard diesel (2017)

lions_city_18C_numbers = [f'4{i:0>2}' for i in range(7,15)] # articulated diesel (2020)
lions_city_g_ng_323_A23_numbers = ([f'26{i:0>1}' for i in range(1,6)] # articulated diesel (2010-2018)
                                   + [f'40{i:0>1}' for i in range(1,6)]
                                   + [f'3{i:0>2}' for i in range(51,65)]
                                   + [f'3{i:0>2}' for i in range(72,84)])
citaro_c2_o530_G_numbers = [f'3{i:0>2}' for i in range(65,72)] # articulated diesel (2017)

eCitaro_numbers = [f'3{i:0>2}' for i in range(27,44)] # standard battery (2020-2021)
eCitaro_G_numbers = [f'4{i:0>2}' for i in range(15,28)] + [f'4{i:0>2}' for i in range(37,47)]  # articulated battery (2021-2024)
lions_city_18E_numbers = [f'4{i:0>2}' for i in range(28,37)] # articulated battery (2024)

electric_bus_numbers = eCitaro_numbers + eCitaro_G_numbers + lions_city_18E_numbers
articlated_bus_numbers = (citaro_c2_o530_G_numbers
                          + lions_city_g_ng_323_A23_numbers
                          + lions_city_18C_numbers
                          + eCitaro_G_numbers
                          + lions_city_18E_numbers)

# colormap to assign (tram)line numbers their official color (as given on the Lininenetzplan) 
# when applied to values ranging from [0,10], where 0 indicates no line at all
heagLineColormap: ListedColormap = ListedColormap(
    ['#ffffff', '#eb6d8c', '#00a650', '#f2bc17', '#3871c1', '#50ade5', '#ec6d1f', '#ec008c', '#e02828', '#74c933',
     '#8d198f'])

# offset to separate numerical line identifiers from utf-8 encodings of other line identifiers
lineId_offset: int = 10000 #TODO: maybe 10^4 better suited?


def encode_line_name(line_name: str) -> int:
    """ encodes the string representation of the line identifier into a unique int

    Args: lineName: the line identifier as a string

    Returns:
        int: the encoded line identifier
    """
    if line_name.isnumeric():
        return int(line_name)

    return int.from_bytes(line_name.encode(), 'big') + lineId_offset


def decode_line_id(line_id: int) -> str:
    """ decodes the sting representation of the line identifier encoded in the lineId

    Args:
        line_id: the encoded line identifier, as given by mapLineName

    Returns:
        str: sting representation of the line name
    """
    if line_id < lineId_offset:
        return str(line_id)

    # line name is encoded
    name_representation = line_id - lineId_offset
    byte_representation = name_representation.to_bytes((name_representation.bit_length() + 7) // 8, 'big')
    return byte_representation.decode('utf-8')



import random
import string
import unittest

from heagVehicleLivedataUtils import vehicleInformation as vi


class TestVehicleNumber(unittest.TestCase):

    def test_prefix_on_trams(self):
        #st10
        for i in range(1, 8 + 1):
            self.assertEqual('76', vi.tram_class_prefix(i))

        # st11
        for i in range(9, 14 + 1):
            self.assertEqual('82', vi.tram_class_prefix(i))

        # st12
        for i in range(15, 24 + 1):
            self.assertEqual('91', vi.tram_class_prefix(i))

        # st13
        for i in range(55, 74 + 1):
            self.assertEqual('98', vi.tram_class_prefix(i))

        # st14
        for i in range(75, 92 + 1):
            self.assertEqual('07', vi.tram_class_prefix(i))

        # st15
        for i in range(101, 125 + 1):
            self.assertEqual('22', vi.tram_class_prefix(i))

        # st7
        self.assertEqual('', vi.tram_class_prefix(31))
        self.assertEqual('', vi.tram_class_prefix(25))
        self.assertEqual('', vi.tram_class_prefix(26))

    def test_tram_number(self):
        #st10
        for i in range(1, 8 + 1):
            self.assertEqual(f'76{i}', vi.get_tram_number(i))

        # st11
        for i in range(9, 14 + 1):
            self.assertEqual(f'82{i}', vi.get_tram_number(i))

        # st12
        for i in range(15, 24 + 1):
            self.assertEqual(f'91{i}', vi.get_tram_number(i))

        # st13
        for i in range(55, 74 + 1):
            self.assertEqual(f'98{i}', vi.get_tram_number(i))

        # st14
        for i in range(75, 92 + 1):
            self.assertEqual(f'07{i}', vi.get_tram_number(i))

        # st15
        for i in range(101, 125 + 1):
            self.assertEqual(f'22{i}', vi.get_tram_number(i))

        # st7
        self.assertEqual('31', vi.get_tram_number(31))
        self.assertEqual('25', vi.get_tram_number(25))
        self.assertEqual('26', vi.get_tram_number(26))

    def test_prefix_other_inputs(self):
        # not tram
        # self.assertRaises(vi.tram_class_prefix(0),ERROR??) TODO: wie wollen wir die handlen?
        # self.assertRaises(vi.tram_class_prefix(-10),ERROR??) TODO: wie wollen wir die handlen?
        # self.assertRaises(vi.tram_class_prefix(433),ERROR_BUS?? not a tram) TODO: wie wollen wir die handlen?

        # assert no prefix
        self.assertEqual('', vi.tram_class_prefix(0))
        self.assertEqual('', vi.tram_class_prefix(-1))
        self.assertEqual('', vi.tram_class_prefix(433))


class TestLineName(unittest.TestCase):

    def test_line_encoding(self):
        # tram lines are encoded as themselves
        for tram_line in range(1, 11):
            self.assertEqual(tram_line, vi.encode_line_name(str(tram_line)))

        # numeric bus lines are also encoded as themselves
        for numeric_bus_line in range(600, 900, 20):
            self.assertEqual(numeric_bus_line, vi.encode_line_name(str(numeric_bus_line)))

        # check regular bus lines
        for _ in range(100):
            line_id = vi.encode_line_name(__random_line_name__())
            self.assertTrue(line_id > 0)

        # test improper inputs
        vi.encode_line_name('')

    def test_line_decoding(self):
        # tram lines are decoded correctly
        for tram_line in range(1, 11):
            self.assertEqual(str(tram_line), vi.decode_line_id(tram_line))
        # numeric bus lines are also decoded correctly

        for numeric_bus_line in range(600, 900, 20):
            self.assertEqual(str(numeric_bus_line), vi.decode_line_id(numeric_bus_line))

        # check regular bus lines
        for _ in range(100):
            line_name = __random_line_name__()
            line_name_decoded = vi.decode_line_id(vi.encode_line_name(line_name))
            self.assertTrue(line_name, line_name_decoded)

        # test improper inputs
        self.assertEqual('', vi.decode_line_id(vi.encode_line_name('')))

        # test decode 0 as no line
        self.assertEqual('No service', vi.decode_line_id(0))


def __random_line_name__(min_length:int =1, max_length:int =10):
    letters = string.ascii_uppercase + string.digits + string.ascii_lowercase
    return ''.join(random.choice(letters) for _ in range(random.randint(min_length, max_length)))

if __name__ == '__main__':
    unittest.main()



"""
TODO: stuff for ploting data
"""

from io import BytesIO

import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

from .data import DataAnalysis
from ..vehicleInformation import decode_line_id, heagLineColormap, modern_tram_numbers, usual_tram_numbers, \
    electric_bus_numbers


def _plot_vehicle_service_timeseries(vehicle_service_timeseries: pd.DataFrame,
                                      *,
                                      filename: str|BytesIO|None = None,
                                      show_plot: bool= True,
                                      show_line_annotation:bool = False,
                                      datespec: str = '%d.%m.%Y-%H:%M',
                                      sample_time: str|None = None,
                                      title: str = 'Vehicles in service',
                                      figsize: tuple[int,int]|None = None,
                                      time_display_factor:float = .2
                                        ):
    """
    plots the timeseries

    Args:
        vehicle_service_timeseries: dataframe containing timeseries with vehicle service
        show_line_annotation: if true, line Names will also be plotted
        datespec: specifies how timestamp is displayed -> used as argument for pandas strftime
        filename: path of file that is used to save the plot. if left empty, nothing will be saved
        show_plot: true if the plot should be shown
        sample_time: sample size of the plot
        figsize: matplotlib figsize to be used for the plot
        title: title of the plot

    """
    if sample_time is not None:
        vehicle_service_timeseries = vehicle_service_timeseries.fillna(0).resample(sample_time).max()

    # fillna a second time to account for missing timeframe
    vehicle_service_timeseries = vehicle_service_timeseries.fillna(0).astype('int64')

    # format index for plotting
    vehicle_service_timeseries.index = vehicle_service_timeseries.index.strftime(datespec)

    if figsize is not None:
        plt.figure(figsize=figsize)
    else:
        number_of_vehicles = vehicle_service_timeseries.shape[1]
        number_of_entries = vehicle_service_timeseries.shape[0]

        plt.figure(figsize=(3+.2*number_of_vehicles,3+ time_display_factor*number_of_entries))

    if show_line_annotation:
        line_annotation = vehicle_service_timeseries.map(lambda x: f'{decode_line_id(x)}' if x != 0 else '')
        sns.heatmap(vehicle_service_timeseries, annot=line_annotation, fmt='', vmin=0, vmax=10,
                    cmap=heagLineColormap, cbar=False)
    else:
        sns.heatmap(vehicle_service_timeseries, vmin=0, vmax=10, cmap=heagLineColormap, cbar=False)

    plt.title(title)

    plt.tight_layout()

    if filename is not None:
        plt.savefig(filename)
    if show_plot:
        plt.show()
    plt.close()

class VehicleDataPlotter (DataAnalysis):
    """
    provides plotting tools
    """

    def plot_number_of_trams_in_service(self,
                                        *,
                                        filename: str|BytesIO|None = None,
                                        show_plot: bool = True,
                                        sample_time: str|None = '10Min',
                                        datespec: str = '%d.%m.%Y-%H:%M',
                                        figsize: tuple[int, int]|None = None,
                                        time_display_factor: float = .2
                                        ):
        """

        Args:
            time_display_factor: scalar for the width of time
            datespec: specififes how timestamp is displayed TODO: was ist da der spec?
            sample_time: sample size of the plot
            figsize:  figsize to be used for the plot
            show_plot: currently only placeholder
            filename: path of file that is used to save the plot. if left empty, nothing will be saved

        """
        number_of_trams_over_time = self.get_number_of_trams_in_service(sample_time)

        if figsize is None:
            number_of_timepoints = number_of_trams_over_time.shape[0]
            figsize = (3 + time_display_factor*number_of_timepoints, 10)

        number_of_trams_over_time.index = number_of_trams_over_time.index.strftime(datespec)

        ax = number_of_trams_over_time.plot(
            kind="bar",
            stacked=True,
            title='HEAG Fahrzeugeinsätzte',
            ylim=(0, 45),
            ylabel="Anzahl Fahrzeuge im Linienbetrieb",
            figsize=figsize,
            width=1)

        plt.tight_layout()

        if filename is not None:
            ax.figure.savefig(filename)
        # TODO: how does this work in the context here?
        #if show_plot:
        #    plt.show()

    def plot_trams_in_servie(self,
                             /,
                             tram_numbers: list[str],
                             *,
                             filename: str|BytesIO|None = None,
                             show_plot: bool= True,
                             show_line_annotation: bool = False, #testing has shown that annotation is what takes most of the runtime
                             datespec: str = '%d.%m.%Y-%H:%M',
                             sample_time: str|None = None,
                             figsize: tuple[int,int]|None = None,
                             time_display_factor:float = .2):
        """
        plots the line assignment for the specifid trams

        Args:
            tram_numbers: list of operational Numbers of the for witch the service status is supposed to be displayed

            time_display_factor: scalar for the height of time
            show_line_annotation: if true, line Names will also be plotted
            datespec: specififes how timestamp is displayed TODO: was ist da der spec?
            filename: path of file that is used to save the plot. if left empty, nothing will be saved
            show_plot: true if the plot should be shown
            sample_time: sample size of the plot
            figsize: figsize to be used for the plot
        """

        _plot_vehicle_service_timeseries(self.get_trams_in_service()['lineid'].reindex(columns=tram_numbers),
                                          title = "Trams in service",
                                          filename=filename,
                                          show_line_annotation=show_line_annotation,
                                          show_plot=show_plot,
                                          datespec=datespec,
                                          sample_time=sample_time,
                                          figsize=figsize,
                                          time_display_factor=time_display_factor)

    def plot_modern_trams_in_service(self,**kwargs):
        """
        like plot_trams_in_servie, but only plots modern trams

        Args:
            Keyword Args: same as plot_trams_in_service except for tram_numbers

        """
        self.plot_trams_in_servie(modern_tram_numbers, **kwargs)

    # das ist eventuell redundent, könnte auch einfach als default wert bei plot_trams_in_servie gemacht werden
    def plot_all_trams_in_service(self, **kwargs):
        """
       like plot_trams_in_servie, but without the ability to specify trams

       Args:
            Keyword Args: same as plot_trams_in_service except for tram_numbers

       """
        self.plot_trams_in_servie(usual_tram_numbers, **kwargs)

    def plot_buses_in_service(self,
                              bus_numbers: list[str],
                               *,
                              filename: str|BytesIO|None = None,
                             show_plot: bool= True,
                             show_line_annotation: bool = True,
                             datespec: str = '%d.%m.%Y-%H:%M',
                             sample_time: str|None = None,
                             figsize: tuple[int,int]|None = None,
                             time_display_factor:float = .2): # TODO columns need to be wider
        """
        plots the line assignment for the electric buses

        Args:
            bus_numbers: list of Betriebsnummber of the buses whose servie is to be plotted

            time_display_factor: scalar for the height of time
            show_line_annotation: if true, line Names will also be plotted
            datespec: specififes how timestamp is displayed TODO: was ist da der spec?
            filename: path of file that is used to save the plot. if left empty, nothing will be saved
            show_plot: true if the plot should be shown
            sample_time: sample size of the plot
            figsize: figsize to be used for the plot
        """

        _plot_vehicle_service_timeseries(self.get_buses_in_service()['lineid'].reindex(columns=bus_numbers),
                                          title="Buses in service",
                                          filename=filename,
                                          show_line_annotation=show_line_annotation,
                                          show_plot=show_plot,
                                          datespec=datespec,
                                          sample_time=sample_time,
                                          figsize=figsize,
                                          time_display_factor=time_display_factor)

    def plot_electric_buses_in_service(self, **kwargs):
        """
        plots the line assignment for the electric buses

        Args:
            Keyword Args: same as plot_buses_in_service except for bus_numbers
        """

        self.plot_buses_in_service(electric_bus_numbers, **kwargs)




"""
provides the DataAnalysis class which provides tools for analyzing the vehicle data

TODO: write more
"""
import pandas as pd

from ..vehicleDataUtils.read import verify_vehicledata_format, vehicledata_from_dir
from ..vehicleDataUtils.process import trams_in_service, number_of_trams_in_service, buses_in_service

# TODO: vllt hier auch weitere optionen?

def _df_changes(df: pd.DataFrame)-> pd.DataFrame:
    """
    returns changes in entires of dataframe

    Args:
        df:

    Returns:
        dataframe of same shape as df, with values True if entries in columns changed. first row will always be True

    """
    return df.ne(df.shift())

def _extract_journeys_for_vehicle(in_service_df:pd.DataFrame, journey_indicators: pd.DataFrame, vehicleid)->pd.DataFrame:
    """
    extracts journeys of specified vehicle
    Args:
        in_service_df: in service dataframe containing information about the vehicle
        journey_indicators: assigns a journey to each row index appearing in in_servie_df. Needs to have a column for vehicleid.
        vehicleid: colum name for the vehicle we want to extract

    Returns:
        dataframe with journeys, with timestamp for begin and end, as well as line number and destination
    """
    time_line_direction_df = in_service_df[[('timestamp', ''), ('lineid',vehicleid), ('direction',vehicleid)]] # in_service_df has mulitinidex columns
    time_line_direction_df.columns = time_line_direction_df.columns.map(lambda x: x[0]) #  removes vehicleid from colum names

    vehicle_journey_indicators = journey_indicators[vehicleid] # we can only group with serial data
    vehicle_journey_indicators.name = 'journey' # we don't want the output index to be named vehicleid

    journeys = time_line_direction_df.groupby(vehicle_journey_indicators)

    return journeys.agg(
        start=pd.NamedAgg(column='timestamp', aggfunc='first'),
        end=pd.NamedAgg(column='timestamp', aggfunc='last'),
        lineid=pd.NamedAgg(column='lineid', aggfunc='first'),
        direction=pd.NamedAgg(column='direction', aggfunc='first')
    )

def _get_vehicle_journeys(in_service_df):
    """

    Args:
        in_service_df: dataaframe formated as returned by trams_in_service

    Returns:
        dataframe with journeys, with timestamp for begin and end, as well as line number and destination.
        index by vehicleid and journeynumber

    """

    # fill empty values, because otherwise NaN != NaN will result in many journeys
    in_service_df['lineid'] = in_service_df['lineid'].fillna(0).astype('int64')
    in_service_df['direction'] = in_service_df['direction'].fillna('')

    # timestamp as column, otherwise it will be lost when grouping
    time_line_destination_df = in_service_df.reset_index()[['timestamp', 'lineid', 'direction']]
    # also needs to happen now, so that the index remains consistent between dataframe and group_indicators

    changes = _df_changes(time_line_destination_df['lineid']) + _df_changes(time_line_destination_df['direction'])

    group_indicators = changes.cumsum() - 1  # first row always indicates a change

    vehicle_ids = time_line_destination_df['lineid'].columns # just grabbing it from here
    vehicle_journeys = [_extract_journeys_for_vehicle(time_line_destination_df, group_indicators, vehicle_id) for
                        vehicle_id in vehicle_ids]

    return pd.concat(vehicle_journeys, axis=0, keys=vehicle_ids)

class DataAnalysis:
    """
    Class to provide tools to analyze given vehicle data.

    TODO: add a bit of detail
    """

    def __init__(self,
                 /,
                 vehicledata: pd.DataFrame = None,
                 *,
                 vehicledata_path: str = None
                 ):
        """


        Args:
            vehicledata: dataframe containing vehicle data. if vehicledata_path or response_paths are specified, this will be disregarded

            vehicledata_path: path to directory where to look for vehicle data json files. if vehicledata is specified, this will be disregarded
        """

        self._vehicledata: pd.DataFrame

        if (vehicledata_path is not None) + (vehicledata is not None) > 1:
            # more than one data sources specified
            Warning("more than one data sources specified, only the fist one will be regarded")


        if vehicledata is not None:
            self.__set_vehicledata_dataframe(vehicledata)
        elif vehicledata_path is not None:
            self.__set_vehicledata_dataframe(vehicledata_from_dir(vehicledata_path))

    def __set_vehicledata_dataframe(self, vehicledata: pd.DataFrame):
        """
        provide vehicledata dataframe directly
        Args:
            vehicledata: dataframe to analyze
        """
        verify_vehicledata_format(vehicledata)

        self._vehicledata = vehicledata

    def get_vehicledata(self) -> pd.DataFrame:
        """

        Returns: the vehicledata dataframe

        """
        return self._vehicledata

    def get_trams_in_service(self, **kwargs) -> pd.DataFrame:
        """
        Args:
            same as process.trams_in_service

        Returns: the dataframe containing the service assignemt of the trams in this analysis

        """
        return trams_in_service(self._vehicledata, **kwargs)

    def get_number_of_trams_in_service(self, sample_time: str|None = None) -> pd.DataFrame:
        """
        Args:
            sample_time: sample size of the vehicledata

        Returns:
            dataframe with numbers of trams in service, index by timestamp.
        """
        number_in_service = number_of_trams_in_service(self.get_trams_in_service()['lineid'])
        if sample_time is not None:
            number_in_service = number_in_service.resample(sample_time).mean()
        return number_in_service

    def get_buses_in_service(self, **kwargs) -> pd.DataFrame:
        """
        Args:
            same as process.buses_in_service

        Returns: the dataframe containing the service assignemt of the buses in this analysis

        """

        return buses_in_service(self._vehicledata, **kwargs)

    def get_tram_journeys(self) -> pd.DataFrame:
        """
        extracts journeys of tram vehicles from vehicle data
        Returns:
            dataframe of journeys indexed by vehicleid and journey count
        """
        return _get_vehicle_journeys(self.get_trams_in_service(columns=['lineid', 'direction']))

    def get_buses_journeys(self) -> pd.DataFrame:
        """
        extracts journeys of buses from vehicle data
        Returns:
            dataframe of journeys indexed by vehicleid and journey count
        """
        return _get_vehicle_journeys(self.get_trams_in_service(columns=['lineid', 'direction']))


"""
methods for working with collection of vehicle data.
"""

import pandas as pd

from ..vehicleInformation import get_tram_number, st12Numbers, st13Numbers, st14Numbers, \
    st15Numbers
from .read import vehicledata_index_timestamp,vehicledata_index_vehicleid

def remove_duplicated_index(df: pd.DataFrame) -> pd.DataFrame:
    """removes entries with duplicated index

    Args:
        df: pandas dataframe

    Returns:
        dataframe where only fist occurrence of each index is kept
        """
    return df[~df.index.duplicated(keep='first')]


def trams_in_service(vehicledata_df: pd.DataFrame, columns:list|None=None) -> pd.DataFrame:
    """ list the lines the trams are in service on, based on the presented vehicleData

    differet methond from busses_in_service because this also adds the prefix to the vehicle id

    Args:
        vehicledata_df: dataFrame as given by vehicleData_from_jsonFiles
        columns: columns to be included in return

    Returns:
        DataFrame: DataFrame indexed by timestamps and columns are the give mulitindexed with trams
    """
    trams = vehicledata_df[vehicledata_df['category'] == 1]

    # remove duplicated index
    trams = remove_duplicated_index(trams)

    if columns is not None:
        trams = trams.reindex(columns=columns)

    if trams.empty:
        raise ValueError('empty Dataframe, cannot pivot')

    trams_in_service_df = trams.reset_index().pivot(index=vehicledata_index_timestamp,
                                                    columns=vehicledata_index_vehicleid)
    trams_in_service_df.columns = trams_in_service_df.columns.map(lambda x: (x[0], get_tram_number(x[1])))

    return  trams_in_service_df


def number_of_trams_in_service(trams_in_service_df: pd.DataFrame) -> pd.DataFrame:
    """
    counts the number of trams in each class, that are in servie for at each timestamp
    TODO: ahh grammar

    Args:
        trams_in_service_df: the service dataframe, formated like return of trams_in_service

    Returns: dataframe containing the counts per class at each timestamp

    """
    tram_class_number_tuples=[("ST12", st12Numbers), ("ST13", st13Numbers), ("ST14", st14Numbers), ("ST15", st15Numbers)]
    number_of_trams_in_service_series_list_by_class: list[pd.Series] = []
    visited_columns: list[str] = []

    for tram_class_tuple in tram_class_number_tuples:
        number_of_trams_of_class_series = (trams_in_service_df.reindex(columns=tram_class_tuple[1])
                                           .fillna(0).map(lambda x: x > 0).sum(axis=1)) # adds one for every tram in service
        number_of_trams_of_class_series.name = tram_class_tuple[0]
        number_of_trams_in_service_series_list_by_class.append(number_of_trams_of_class_series)

        visited_columns.extend(tram_class_tuple[1])

    other_trams = trams_in_service_df.columns.difference(visited_columns)
    number_of_other_trams_in_service = trams_in_service_df.reindex(columns=other_trams).fillna(0).map(lambda x: x > 0).sum(axis=1)
    number_of_other_trams_in_service.name="other"
    number_of_trams_in_service_series_list_by_class.append(number_of_other_trams_in_service)

    return pd.concat(number_of_trams_in_service_series_list_by_class,axis=1)


def buses_in_service(vehicledata_df: pd.DataFrame, columns:list|None=None) -> pd.DataFrame:
    """ list the lines the buses are in service on, based on the presented vehicleData

    Args:
        vehicledata_df (DataFrame): dataFrame as given by vehicleData_from_jsonFiles
        columns: columns to be included in return

    Returns:
        DataFrame: DataFrame indexed by timestamps that contains the line the selected buses are in service on
    """
    buses = vehicledata_df[vehicledata_df['category'] == 5]

    buses = remove_duplicated_index(buses)

    if columns is not None:
        buses = buses.reindex(columns=columns)

    if buses.empty:
        raise ValueError('empty Dataframe, cannot pivot')
    buses_in_service_df = buses.reset_index().pivot(index=vehicledata_index_timestamp,
                                                    columns=vehicledata_index_vehicleid)
    buses_in_service_df.columns = buses_in_service_df.columns.map(lambda x: (x[0],str(x[1]))) # make sure that we index with stings

    return  buses_in_service_df



"""
Methods for reading vehiclelivedata from json
"""
from os import walk
from os.path import join

from pathlib import Path

import pandas as pd
import json

from ..vehicleInformation import encode_line_name

vehicledata_columns = ["lineid", "category", "direction", "status", "latitude", "longitude", "bearing", "type"]
vehicledata_index_timestamp = 'timestamp'
vehicledata_index_vehicleid = 'vehicleid'
vehicledata_index_names = [vehicledata_index_timestamp, vehicledata_index_vehicleid]

def verify_vehicledata_format(dataframe: pd.DataFrame) -> bool:
    """
    checks if dataframe contains a valid vehicle data format
    Args:
        dataframe: dataframe to check

    Throws: Value error if dataframe is not formatted correctly

    """

    expected_columns = set(vehicledata_columns) # is set so that order does not matter -> TODO: is that necessary?
    if not expected_columns <= set(dataframe.columns):
        raise ValueError("dataframe columns do not contain expected columns")

    expected_index_names = set(vehicledata_index_names)
    if not expected_index_names == set(dataframe.index.names):
        raise ValueError("dataframe index names do not match expected names")

    return True

## data reading
def vehicledata_from_dict(vehicledata_dict: dict, *, file_path: str = None)-> pd.DataFrame:
    """ extracts service information of public transport vehicles into a Dataframe

    Args:
        vehicledata_dict (JSON dict): data structured like vehicleData from HEAG vehicleLivedata api
        file_path: used in error messages. If vehicledata_dict is taken from a json file, use its path, otherwise leave empty

    Returns:
        DataFrame: contains the information from the vehicleData, indexed with timestamp and vehicleId
    """
    if not {'timestamp', 'vehicles'} <= vehicledata_dict.keys():
        raise ValueError(
            f"json file {file_path} is not formatted correctly. Missing keys 'timestamp' and/or 'vehicles' ")

    vehicledata_df = pd.DataFrame.from_dict(vehicledata_dict['vehicles'])

    # lowercase colums work better with database
    vehicledata_df.columns = vehicledata_df.columns.str.lower()

    expected_entries_names = set(vehicledata_columns + [vehicledata_index_vehicleid])
    if not expected_entries_names <= set(vehicledata_df.columns):
        if file_path is None:
            message_begin = "vehicledata_dict"
        else:
            message_begin = f"json file {file_path}"

        missing_entries = expected_entries_names - set(vehicledata_df.columns)
        raise ValueError(message_begin + f"was not formatted correctly. Missing entries {missing_entries} in vehicles")

    # use timestamp vehicleId multiindex -> TODO: überlege ob das sinvoll ist (bei db sollte alles in colums stehen, sonnst ist vllt anders praktisch), ist jetzt aba alles darauf ausgelegt
    vehicledata_df.index = pd.MultiIndex.from_product(
                                            [[pd.Timestamp(vehicledata_dict['timestamp'])],
                                                    vehicledata_df[vehicledata_index_vehicleid]],
                                                    names= vehicledata_index_names )

    vehicledata_df['lineid'] = vehicledata_df['lineid'].map(encode_line_name)

    # make sure columns are the expected ones
    vehicledata_df = vehicledata_df.reindex(columns = vehicledata_columns)

    return vehicledata_df

# TODO: | PathLike[str] as well??
def vehicledata_from_dir(path_to_dir: str|Path, max_recursion_depht: int =0, follow_links:bool = False) -> pd.DataFrame:
    """
    returns the vehicledata found in dir
    Args:
        path_to_dir: where to search for vehicledata
        max_recursion_depht: deprecated

    Returns:
        vehicledata found in dir, returned as dataframe
    """

    vehicledata_df_list = []
    for dirpath, dirname, filenames in walk(path_to_dir,follow_links):
        for filename in filenames:
            with open(join(dirpath, filename)) as json_file:
                vehicledata_df_list.append(vehicledata_from_dict(json.load(json_file)))

    if len(vehicledata_df_list) == 0:
        raise ValueError("No vehicledata found in directory")

    return pd.concat(vehicledata_df_list)

